#------------------------------------------------------------------
# Copyright (c) 2006 Symantec Corporation. All rights reserved.
#
# Symantec,  the  Symantec  Logo,  Veritas,  and   Veritas  Storage
# Foundation   are   trademarks   or   registered   trademarks   of
# Symantec Corporation  or  its affiliates in the  U.S.  and  other
# countries.  Other  names  may  be  trademarks of their respective
# owners.
#
# The Licensed Software  and  Documentation  are   deemed   to   be
# "commercial   computer   software"    and   "commercial  computer
# software documentation"  as defined in FAR  Sections  12.212  and
# DFARS Section 227.7202.
#
# $Header: /project/isv-agent/VProAgents/SAPHotStandby/SAPHotStandby.pm,v 1.8 2011/08/22 10:11:29 lpatil Exp $
#
# $Author: lpatil $
# $Date: 2011/08/22 10:11:29 $
# $Revision: 1.8 $
#
# Provides:                            SAPHotStandby
# Author:                              vchappav
# CrAgent Run Date:                    09 Mar, 2007
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package SAPHotStandby;

use IPC::SysV qw ( IPC_STAT IPC_RMID );

use Sys::Hostname;

use constant STANDBY        => 200;
use constant ONLINE         => 300;
use constant INVALID        => 400;

my $sCmdIpcs   = "/usr/bin/ipcs";

my ( $CLUS_HOME, $sPrefix ) = ();

#---------------------
# Setup VCS/VAD home..
#---------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.0/VAD..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvad";
}
else { # VCS
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /(...)$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = "$CLUS_HOME/bin/hares";
my $sCmdHagrp  = "$CLUS_HOME/bin/hagrp";
my $sCmdHatype = "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

my @liRunningProcs = (); # Global..

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses
use SystemD;        # Support generalized SystemD Functions

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use POSIX qw ( uname );
use POSIX qw ( strftime );
use File::Path qw ( mkpath );
use File::stat;

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: SAPHotStandby.pm,v 1.8 2011/08/22 10:11:29 lpatil Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define SAPHotStandby subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String EnvFile,
                              IN String LCHome,
                              IN String LCHost,
                              IN String LCUser,
                              IN String LCName,
                              IN String DBMUserId,
                              IN String DBMPassword,
                              IN String DBAUserId,
                              IN String DBAPassword,
                              IN String StandbyList,
                              IN String StandbyRole,
                              IN String MasterResName,
                              IN String MonitorProgram );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localEnvFile                  Environments file for 'LCUser'. 
 $localLCHome                   Directory for liveCache/MaxDB server software.
 $localLCHost                   Virtual Host Name for liveCache/MaxDB server.
 $localLCUser                   Operating system user for liveCache/MaxDB server.
 $localLCName                   Name of the liveCache/MaxDB instance.
 $localDBMUserId                Database user ID used to connect to liveCache/MaxDB server.
 $localDBMPassword              Password for the user 'DBMUserId'.
 $localDBAUserId                Database Admin user ID used to query liveCache/MaxDB server.
 $localDBAPassword              Password for the user 'DBAUserId'.
 $localStandbyList              List of Standby nodes where standby startup is needed.
 $localStandbyRole              Role of the liveCache/MaxDB server either MASTER or STANDBY.
 $localMasterResName            Name of the VCS resource for liveCache/MaxDB MASTER server.
 $localMonitorProgram           Full path to external monitor program.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localEnvFile,
       $localLCHome,
       $localLCHost,
       $localLCUser,
       $localLCName,
       $localDBMUserId,
       $localDBMPassword,
       $localDBAUserId,
       $localDBAPassword,
       $rlsStandbyList,
       $localStandbyRole,
       $localMasterResName,
       $localMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCHost                   [%s]
  LCUser                   [%s]
  LCName                   [%s]
  DBMUserId                [%s]
  DBMPassword              [%s]
  DBAUserId                [%s]
  DBAPassword              [%s]
  StandbyList              [%s]
  StandbyRole              [%s]
  MasterResName            [%s]
  MonitorProgram           [%s]",

  $sMyName,    $sEntryPointName,    $localEnvFile,    $localLCHome,
  $localLCHost,    $localLCUser,   $localLCName,   $localDBMUserId,
  $localDBMPassword,      $localDBAUserId,       $localDBAPassword,
  join(chr 040, @$rlsStandbyList),     $localStandbyRole,    
  $localMasterResName,         $localMonitorProgram );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName EnvFile  LCHome  LCHost  LCUser  LCName  DBMUserId
       DBMPassword   DBAUserId   DBAPassword   StandbyList   StandbyRole
       MasterResName MonitorProgram );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( EnvFile LCHome LCHost LCUser LCName DBMUserId DBMPassword DBAUserId DBAPassword StandbyRole );
  my @lsOfflineArgs = qw ( EnvFile LCHome LCHost LCUser LCName DBMUserId DBMPassword StandbyRole );
  my @lsMonitorArgs = qw ( EnvFile LCHome LCHost LCUser LCName DBMUserId DBMPassword StandbyRole );
  my @lsCleanArgs   = qw ( EnvFile LCHome LCHost LCUser LCName DBMUserId DBMPassword StandbyRole );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{$hlsNeeded { $sEntryPointName }};

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------------------
  # Individual validation starts..
  #-------------------------------

  #-----------------------------------------------
  # Validation for online offline entry point(s)..
  #-----------------------------------------------
  if ( $sEntryPointName =~ /^(online|offline)$/o ) {
    #-----------------------------
    # Validate parameter EnvFile..
    #-----------------------------
    if ( ! -f  $localEnvFile ) {
      &logError ( 20003, "<%s> EnvFile [%s] is not a plain file",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }

    #----------------------------
    # Validate parameter LCHome..
    #----------------------------
    if ( ! -d  $localLCHome ) {
      &logError ( 20004, "<%s> LCHome [%s] directory does not exist",
      $sMyName, $localLCHome );
      $iReturnCode = 0; # Failure..
    }

    #--------------------------
    # Validate parameter LCUser..
    #--------------------------
    if ( ! defined getpwnam $localLCUser ) {
      &logError ( 20005, "<%s> LCUser [%s] is not a valid UNIX login",
      $sMyName, $localLCUser );
      $iReturnCode = 0; # Failure..
    }

    #--------------------------
    # Validate parameter LCHost..
    #--------------------------
    if ( ! inet_aton $localLCHost ) {
      &logError ( 20006, "<%s> LCHost [%s] is an invalid Hostname/IPv4 address",
      $sMyName, $localLCHost );
      $iReturnCode = 0; # Failure..
    }
  }

  #----------------------------------
  # Validation for all entry points..
  #----------------------------------

  #-----------------------------
  # Validate parameter LCName..
  #-----------------------------
  if ( $localLCName !~ /(^[A-Z][A-Z0-9][A-Z0-9]$)/ ) {
    &logError ( 20007, "<%s> LCName [%s] contains invalid characters",
    $sMyName, $localLCName );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------------------
  # Validate parameter StandbyRole ..
  #-----------------------------------
  my @lsStandbyRoles = qw ( master standby );

  unless ( grep { /^$localStandbyRole$/io } @lsStandbyRoles ) {
    &logError ( 20008,
    "<%s> Standby Role is [%s] invalid. Has to be one of [%s]",
    $sMyName, $localStandbyRole, join ( ' ', @lsStandbyRoles ) );
    $iReturnCode = 0; # Failure..
  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20009, "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }

  }


  #-----------------------------------------------------------------------------
  # Validate StandbyList for all the Standbys  for online and monitor ...
  #-----------------------------------------------------------------------------
  my @lsStandbyList = @$rlsStandbyList;  # Backup ..

  if ( scalar @lsStandbyList && $sEntryPointName =~ /^(online|monitor)$/o ) {
    foreach my $sStandby ( @lsStandbyList ) {
      &logTrace( 20010, "<%s> Got HotStandby node name [%s]", $sMyName, $sStandby );
      #----------------------------
      # Validate Standby Server ..
      #----------------------------
      if ( $sStandby ne "" && ! inet_aton ( $sStandby ) ) {
        &logError ( 20011, "<%s> Standby Server [%s] is invalid",
        $sMyName, $sStandby );
        if( $iReturnCode ) {
          $iReturnCode = 0; # Failure..
        }
      }
    }
  }


  $iReturnCode ||
  &logError ( 20012,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String DBMPassword,
                         IN String DBMUserId,
                         IN String EnvFile,
                         IN String LCHome,
                         IN String LCHost,
                         IN String LCName,
                         IN String LCUser,
                         IN String StandbyRole,
                         IN Bool   CleanIPC );

 Description : Start logic needed for online entry pt

 Parameters :

 $sDBMPassword              Password for the user 'DBMUserId'.
 $sDBMUserId                Database user ID used to connect to liveCache/MaxDB server.
 $sEnvFile                  Environments file for 'LCUser'. 
 $sLCHome                   Directory for liveCache/MaxDB server software.
 $sLCHost                   Virtual Host Name for liveCache/MaxDB server.
 $sLCName                   Name of the liveCache/MaxDB instance.
 $sLCUser                   Operating system user for liveCache/MaxDB server.
 $sStandbyRole              Role for this instance of liveCache server.
 $bCleanIPC                 Flag to indicate whether IPCs for liveCache needs to be cleaned before start.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBMPassword,
       $sDBMUserId,
       $sEnvFile,
       $sLCHome,
       $sLCHost,
       $sLCName,
       $sLCUser,
       $sStandbyRole,
       $bCleanIPC ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20013, "Subroutine <%s> called with:
  DBMPassword              [%s]
  DBMUserId                [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCHost                   [%s]
  LCName                   [%s]
  LCUser                   [%s]
  StandbyRole              [%s]
  CleanIPC                 [%s]",

  $sMyName,  $sDBMPassword,   $sDBMUserId,   $sEnvFile,   $sLCHome,
  $sLCHost, $sLCName, $sLCUser, $sStandbyRole, $bCleanIPC );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  if( $bCleanIPC ) {
    #-------------------------------------------------
    # Whack any preexisting SAPHotStandby processes..
    #-------------------------------------------------
    &logTrace ( 20014,
    "<%s> Killing SAPHotStandby processes for this instance", $sMyName );

    if ( SAPHotStandby::KillSAPHotStandbyProcesses ( $sLCName,
                                                     $sLCUser ) ) {
      &logTrace ( 20015,
      "<%s> Processes for this instance do not exist before start-up",
      $sMyName );
    }
    else {
      &logError ( 20016,
      "<%s> Could not remove all processes for this instance. Exiting",
      $sMyName );
      $iReturnCode = 0; # Failure..
    }

    #--------------------
    # Return, if failed..
    #--------------------
    $iReturnCode || return $iReturnCode;

    #------------------------------------------------------
    # Remove any preexisting SAPHotStandby IPC resources...
    #------------------------------------------------------
    &logTrace ( 20017,
    "<%s> Cleaning SAPHotStandby IPC resources for instance [%s]", $sMyName, $sLCName );

    if( SAPHotStandby::RemoveIPCResourcesForSAPHotStandby ( "sdb",
                                                            $sLCHome,
                                                            $sLCName ) ) {
      &logTrace(20018,
      "<%s> Removed all IPC resources for SAP HotStandby Instance [%s] successfully",
      $sMyName, $sLCName );
    }
    else {
      &logError(20019,
      "<%s> Removed all IPC resources for SAP HotStandby Instance [%s] successfully",
      $sMyName, $sLCName );
      $iReturnCode = 0; # Failure ..
    }

    #--------------------
    # Return, if failed..
    #--------------------
    $iReturnCode || return $iReturnCode;
  }


  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPHotStandby::GetEngineTimeout ( 1 );
  my ( @lsCmdout, $iExitCode, $iTimeout, @liPids ) = ();

  #-------------------------------------------
  # Start x_server for SAP HotStandby Server..
  #-------------------------------------------
  my $sCommand = "$sLCHome/x_server";
  my $sCommandArgs = "start";

  &logTrace ( 20020,
  "<%s> Starting x_server with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20021,
    "<%s> Insufficient time to start x_server. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = 0; # Failure..
  }
  elsif( $iReturnCode = SystemD::DetectSystemd() ) { 
      my $WhichCommand = "XServer";
      if( $iReturnCode = SystemD::RunSystemdWithEnv($sLCUser, $sCommand, $sCommandArgs, $sEnvFile, $iTimeout, $WhichCommand) ){
        &logTrace ( 20022,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
      }    
  }
  elsif ( $iReturnCode =
    SAPHotStandby::ExecLocalCmd ( $sLCUser,
                                  $sEnvFile,
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20022,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    my $sSrchStr = qr|\bXSERVER\b.*\bstarted\b|io; # Search pattern..
    my $sSrchStr1 = qr|\bFound\s+running\s+XServer\b|io; # Search pattern..
    my $sSrchStr2 = qr|\bFound\s+running.*listener|io; # Search pattern.. 
    # [2016-09-17 20:21:09 22579 89280 INF  12916          Found running sdbgloballistener on port 7210
    #  2016-09-17 20:21:09 22579 89280 INF  13011            version 'U64/LIX86 7.9.08   Build 035-123-254-630' ]

    &logTrace ( 20023, "<%s> Looking for pattern [%s] or [%s] or [%s]  in output",
    $sMyName, $sSrchStr , $sSrchStr1, $sSrchStr2);

    #--------------------------------------------
    # Attempt to locate search string in output..
    #--------------------------------------------
    unless ( grep { /$sSrchStr/ || /$sSrchStr1/ || /$sSrchStr2/ } @lsCmdout ) {
      &logError ( 20024,
      "<%s> Failed to locate search pattern [%s] or [%s] or [%s] in Cmd output [%s]",
      $sMyName, $sSrchStr, $sSrchStr1, $sSrchStr2, join ( chr 012, @lsCmdout ) );

      $iReturnCode = 0; # Failure..
    } else {
      &logTrace ( 20024,
      "<%s> Located search pattern [%s] or [%s] or [%s] in Cmd output [%s]",
      $sMyName, $sSrchStr, $sSrchStr1, $sSrchStr2, join ( chr 012, @lsCmdout ) );
    }
  }
  else {
    &logError ( 20025,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );
    $iReturnCode = 0; # Failure..
  }

  #-------------------
  # Return if failed..
  #-------------------
  &logTrace ( 20024, "<%s> ReturnCode is [%s] ", $sMyName, $iReturnCode );
  $iReturnCode || return $iReturnCode;
  

  #-------------------------------------------
  # Declare start command and args variables..
  #-------------------------------------------
  $sCommand = "$sLCHome/dbmcli";

  if( $sStandbyRole =~ /^master$/io ) {
    $sCommandArgs = "-d $sLCName -n $sLCHost -u $sDBMUserId,\'$sDBMPassword\' db_online";
  }
  elsif( $sStandbyRole =~ /^standby$/io ) {
    #----------------------------
    # Get the local hostname ...
    #----------------------------
    my $sHostname  = hostname;
    my @lsHostname = split( /./, $sHostname, 1 );
    my $localHost  = $lsHostname[0];
    chomp $localHost;

    $sCommandArgs = "-d $sLCName -n $sLCHost -u $sDBMUserId,\'$sDBMPassword\' db_standby $localHost";
  }
  else {
    &logError ( 20026,
    "<%s> Invalid StandbyRole [%s]",
    $sMyName, $sStandbyRole );

    $iReturnCode = 0;  # Failure..
  }

  ( @lsCmdout, $iExitCode, $iTimeout, @liPids ) = ();

  &logTrace ( 20027,
  "<%s> Starting SAPHotStandby instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20028,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif( $iReturnCode = SystemD::DetectSystemd() ) { 
      my $WhichCommand = "Dbmcli";
      if( $iReturnCode = SystemD::RunSystemdWithEnv($sLCUser, $sCommand, $sCommandArgs, $sEnvFile, $iTimeout, $WhichCommand,$sDBMPassword) ){
        &logTrace ( 20022,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
      }    
  }
  elsif ( $iReturnCode =
    SAPHotStandby::ExecLocalCmd ( $sLCUser, 
                                  $sEnvFile, 
                                  $sCommand, 
                                  $sCommandArgs, 
                                  \$iExitCode, 
                                  $iTimeout, 
                                  \@lsCmdout, 
                                  $sDBMPassword ) ) {
    &logTrace ( 20029,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logError ( 20030,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]. 
    Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  my $iLCState = undef;
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20031,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout );

    #---------
    # Do FLM..
    #---------
    $iLCState = SAPHotStandby::GetResourceState ( $sDBMPassword,
                                                  $sDBMUserId,
                                                  $sEnvFile,
                                                  $sLCHome,
                                                  $sLCHost,
                                                  $sLCName,
                                                  $sLCUser,
                                                  $sStandbyRole );

    if ( VCSDef_MonitorExit::ONLINE == $iLCState || 
         SAPHotStandby::ONLINE      == $iLCState || 
         SAPHotStandby::STANDBY     == $iLCState ) { 
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20032,
      "<%s> Confirmed start-up of this instance of SAPHotStandby server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20033,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0;  # Failure..
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String DBMPassword,
                        IN String DBMUserId,
                        IN String EnvFile,
                        IN String LCHome,
                        IN String LCHost,
                        IN String LCName,
                        IN String LCUser,
                        IN String StandbyRole );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sDBMPassword              Password for the user 'DBMUserId'.
 $sDBMUserId                Database user ID used to connect to liveCache/MaxDB server.
 $sEnvFile                  Environments file for 'LCUser'. 
 $sLCHome                   Directory for liveCache/MaxDB server software.
 $sLCHost                   Virtual Host Name for liveCache/MaxDB server.
 $sLCName                   Name of the liveCache/MaxDB instance.
 $sLCUser                   Operating system user for liveCache/MaxDB server.
 $sStandbyRole              Role for this instance of liveCache server.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBMPassword,
       $sDBMUserId,
       $sEnvFile,
       $sLCHome,
       $sLCHost,
       $sLCName,
       $sLCUser,
       $sStandbyRole ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20034, "Subroutine <%s> called with:
  DBMPassword              [%s]
  DBMUserId                [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCHost                   [%s]
  LCName                   [%s]
  LCUser                   [%s]
  StandbyRole              [%s]",

  $sMyName,  $sDBMPassword,   $sDBMUserId,   $sEnvFile,   $sLCHome,
  $sLCHost, $sLCName, $sLCUser, $sStandbyRole );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPHotStandby::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  
  $sCommand = "$sLCHome/dbmcli";

  if( $sStandbyRole =~ /^master$/io ) {
    $sCommandArgs = "-d $sLCName -n $sLCHost -u $sDBMUserId,\'$sDBMPassword\' db_offline";
  }
  elsif( $sStandbyRole =~ /^standby$/io ) {
    #----------------------------
    # Get the local hostname ...
    #----------------------------
    my $sHostname  = hostname;
    my @lsHostname = split( /./, $sHostname, 1 );
    my $localHost  = $lsHostname[0];
    chomp $localHost;

    $sCommandArgs = "-d $sLCName -n $sLCHost -u $sDBMUserId,\'$sDBMPassword\' hss_execute $localHost db_offline";
  }
  else {
    &logError ( 20035,
    "<%s> Invalid StandbyRole [%s]",
    $sMyName, $sStandbyRole );

    $iReturnCode = 0;  # Failure..
  }

  &logTrace ( 20036,
  "<%s> Stopping SAPHotStandby with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20037,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          SAPHotStandby::ExecLocalCmd ( $sLCUser, 
                                        $sEnvFile, 
                                        $sCommand, 
                                        $sCommandArgs, 
                                        \$iExitCode, 
                                        $iTimeout, 
                                        \@lsCmdout, 
                                        $sDBMPassword ) ) {
    &logTrace ( 20038,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20039,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      SAPHotStandby::GetProcState ( \@liPids,
                                    $sLCName,
                                    $sLCUser,
                                    0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20040,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout );
      }
      else {
        &logInfo ( 20041,
        "<%s> Confirmed shutdown of SAPHotStandby instance",
        $sMyName );
        last; # Get out..
      }
    }
  }
  else {
    &logError ( 20042,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );

    $iReturnCode = 0; # Failure ..
  }

  #-----------------------------------------
  # Whack any postexisting SAPHotStandby processes..
  #-----------------------------------------
  &logTrace ( 20043,
  "<%s> Killing SAPHotStandby processes for this instance", $sMyName );

  if ( SAPHotStandby::KillSAPHotStandbyProcesses ( $sLCName,
                                                   $sLCUser ) ) {
    &logTrace ( 20044,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
    $iReturnCode = 1; # Success ...
  }
  else {
    &logError ( 20045,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  #------------------------------------------------------
  # Remove any left over SAPHotStandby IPC resources...
  #------------------------------------------------------
  &logTrace ( 20046,
  "<%s> Cleaning SAPHotStandby IPC resources for instance [%s]", $sMyName, $sLCName );

  if( SAPHotStandby::RemoveIPCResourcesForSAPHotStandby ( "sdb",
                                                          $sLCHome,
                                                          $sLCName ) ) {
    &logTrace (20047,
               "<%s> Removed all IPC resources for SAP HotStandby Instance [%s] successfully",
               $sMyName, $sLCName );
  }
  else {
    &logError (20048,
               "<%s> Removed all IPC resources for SAP HotStandby Instance [%s] successfully",
               $sMyName, $sLCName );
    $iReturnCode = 0; # Failure ..
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String DBMPassword,
                        IN String DBMUserId,
                        IN String EnvFile,
                        IN String LCHome,
                        IN String LCHost,
                        IN String LCName,
                        IN String LCUser,
                        IN String StandbyRole );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sDBMPassword              Password for the user 'DBMUserId'.
 $sDBMUserId                Database user ID used to connect to liveCache/MaxDB server.
 $sEnvFile                  Environments file for 'LCUser'. 
 $sLCHome                   Directory for liveCache/MaxDB server software.
 $sLCHost                   Virtual Host Name for liveCache/MaxDB server.
 $sLCName                   Name of the liveCache/MaxDB instance.
 $sLCUser                   Operating system user for liveCache/MaxDB server.
 $sStandbyRole              Role for this instance of liveCache server.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sDBMPassword,
       $sDBMUserId,
       $sEnvFile,
       $sLCHome,
       $sLCHost,
       $sLCName,
       $sLCUser,
       $sStandbyRole ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20049, "Subroutine <%s> called with:
  CleanReason              [%s]
  DBMPassword              [%s]
  DBMUserId                [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCHost                   [%s]
  LCName                   [%s]
  LCUser                   [%s]
  StandbyRole              [%s]",

  $sMyName, $sCleanReason, $sDBMPassword,  $sDBMUserId,  $sEnvFile,
  $sLCHome, $sLCHost, $sLCName, $sLCUser, $sStandbyRole );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPHotStandby::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  

  $sCommand = "$sLCHome/dbmcli";

  if( $sStandbyRole =~ /^master$/io ) {
    $sCommandArgs = "-d $sLCName -n $sLCHost -u $sDBMUserId,\'$sDBMPassword\' db_offline";
  }
  elsif( $sStandbyRole =~ /^standby$/io ) {
    &logWarn ( 20050,
    "<%s> Skipping clean for STANDBY server.", $sMyName );

    return $iReturnCode; # Don't execute clean utility for STANDBY.. 
  }
  else {
    &logError ( 20051,
    "<%s> Invalid StandbyRole [%s]", $sMyName, $sStandbyRole );

    $iReturnCode = 0;  # Failure..
  }


  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20052,
    "<%s> Attempting to stop SAPHotStandby with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20053,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode = SAPHotStandby::ExecLocalCmd ( $sLCUser, 
                                                         $sEnvFile, 
                                                         $sCommand, 
                                                         $sCommandArgs, 
                                                         \$iExitCode, 
                                                         $iTimeout, 
                                                         \@lsCmdout, 
                                                         $sDBMPassword ) ) {
      &logTrace ( 20054, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      &logError ( 20055,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
      Cmd output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode,
      join ( chr 012, @lsCmdout ) );
    }
  }

  #-----------------------------------------
  # Whack any preexisting SAPHotStandby processes..
  #-----------------------------------------
  &logTrace ( 20056,
  "<%s> Killing SAPHotStandby processes for this instance", $sMyName );

  if ( SAPHotStandby::KillSAPHotStandbyProcesses ( $sLCName,
                                                   $sLCUser ) ) {
    &logTrace ( 20057,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20058,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #-------------------------------------------
  # Define clear command and args variables..
  #-------------------------------------------
  $sCommand = "$sLCHome/dbmcli";
  $sCommandArgs = "-d $sLCName -n $sLCHost -u $sDBMUserId,\'$sDBMPassword\' db_clear";

  #-------------------
  # Try to be gentle..
  #-------------------
  &logInfo ( 20059,
  "<%s> Attempting to clear SAPHotStandby with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #-----------------
  # Request clear ..
  #-----------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20060,
    "<%s> Insufficient time to continue clear. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    #$iReturnCode = 0; # Failure..
  }
  elsif ( SAPHotStandby::ExecLocalCmd( $sLCUser,
                                       $sEnvFile,
                                       $sCommand,
                                       $sCommandArgs,
                                       \$iExitCode,
                                       $iTimeout,
                                       \@lsCmdout,
                                       $sDBMPassword ) ) {
    &logTrace ( 20061,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #------------------------------------------------------
  # Remove any left over SAPHotStandby IPC resources...
  #------------------------------------------------------
  &logTrace ( 20062,
  "<%s> Cleaning SAPHotStandby IPC resources for instance [%s]", $sMyName, $sLCName );

  if( $iReturnCode = SAPHotStandby::RemoveIPCResourcesForSAPHotStandby ( "sdb",
                                                                         $sLCHome,
                                                                         $sLCName ) ) {
    &logTrace(20063,
    "<%s> Removed all IPC resources for SAP HotStandby Instance [%s] successfully",
    $sMyName, $sLCName );
  }
  else {
    &logError(20064,
    "<%s> Failed to Remove all IPC resources for SAP HotStandby Instance [%s]",
    $sMyName, $sLCName );

    $iReturnCode = 0; # Failure ..
  }  
  

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String DBMPassword,
                                   IN String DBMUserId,
                                   IN String EnvFile,
                                   IN String LCHome,
                                   IN String LCHost,
                                   IN String LCName,
                                   IN String LCUser,
                                   IN String StandbyRole );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sDBMPassword              Password for Database Manager operator user.
 $sDBMUserId                User name for Database Manager operator.
 $sEnvFile                  Environments file for the liveCache server.
 $sLCHome                   Home directory for the liveCache server.
 $sLCHost                   Virtual hostname for the liveCache server.
 $sLCName                   Name of the liveCache/MaxDB instance.
 $sLCUser                   Operating system user for liveCache/MaxDB server.
 $sStandbyRole              Role for this instance of liveCache server.
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBMPassword,
       $sDBMUserId,
       $sEnvFile,
       $sLCHome,
       $sLCHost,
       $sLCName,
       $sLCUser,
       $sStandbyRole ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20065, "Subroutine <%s> called:
  DBMPassword              [%s]
  DBMUserId                [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCHost                   [%s]
  LCName                   [%s]
  LCUser                   [%s]
  StandbyRole              [%s]",

  $sMyName, $sDBMPassword, $sDBMUserId, $sEnvFile, $sLCHome, $sLCHost,
  $sLCName, $sLCUser, $sStandbyRole );

  my @liDummyPids = ();
  my $iReturnCode = SAPHotStandby::GetProcState ( \@liDummyPids,
                                                  $sLCName,
                                                  $sLCUser,
                                                  1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20066,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20067,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );

    #--------------------------------
    # Populate the global PID array..
    #--------------------------------
    @liRunningProcs = @liDummyPids;

    return ( $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;

  &logTrace ( 20068,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  #-----------------------------
  # Check the database state ..
  #-----------------------------

  #--------------------------
  # Just In Time Validation..
  #--------------------------
  if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
       SAPHotStandby::DelayedAttributeCheck ( $sDBMPassword,
                                              $sDBMUserId,
                                              $sEnvFile,
                                              $sLCHome,
                                              $sLCHost,
                                              $sLCUser ) ) ) {
    &logTrace ( 20069,
    "<%s> Attribute check completed successfully",
    $sMyName );
  }
  else {
    &logError ( 20070,
    "<%s> Attribute validation failed. Returning [%s]",
    $sMyName, $iReturnCode );

    return $iReturnCode;
  }

  &logTrace ( 20071,
  "<%s> Proceeding with database state check", $sMyName );

  #-------------------------
  # Define local variables..
  #-------------------------
  my $iTillWhen = undef;
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();

  my ( $sEntryPointName ) = ();
  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sCommand = "$sLCHome/dbmcli";
  $sCommandArgs = "-d $sLCName -u $sDBMUserId,\'$sDBMPassword\' db_state";

  #-------------------------------------------------------
  # Upper bound..
  # Time in EPOCH when entry pt started + VCS time limit..
  #-------------------------------------------------------
  $iTillWhen = $^T + SAPHotStandby::GetEngineTimeout ( 1 );

  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20072,
    "<%s> Insufficient time to continue clear. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    #$iReturnCode = 0; # Failure..
  }
  elsif ( SAPHotStandby::ExecLocalCmd ( $sLCUser,
                                        $sEnvFile,
                                        $sCommand,
                                        $sCommandArgs,
                                        \$iExitCode,
                                        $iTimeout,
                                        \@lsCmdout,
                                        $sDBMPassword ) ) {
    &logTrace ( 20073,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( $lsCmdout [-1] eq 'ONLINE' ) {
      &logTrace ( 20074,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      if ( $sEntryPointName eq 'online' && $sStandbyRole =~ /^master$/io ) {
        &logTrace ( 20075,
        "<%s> Database is online", $sMyName );
        $iReturnCode = SAPHotStandby::ONLINE;
      }
      elsif ( $sEntryPointName eq 'online' && $sStandbyRole =~ /^standby$/io ) {
        # Need to prevent this scenario with preonline trigger. :)
        &logWarn ( 20076,
        "<%s> Database state on the node is ONLINE.", $sMyName );

        $iReturnCode = SAPHotStandby::INVALID;
      }
      elsif ( $sStandbyRole =~ /^master$/io ) {
        $iReturnCode = VCSDef_MonitorExit::ONLINE;
      }
      elsif ( $sStandbyRole =~ /^standby$/io ) {
        $iReturnCode = VCSDef_MonitorExit::OFFLINE;
      }
    }
    elsif ( $lsCmdout [-1] eq 'STANDBY' ) {
      &logTrace( 20077,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      #----------------------------------------------
      # Need to test this block of code carefully ..
      #----------------------------------------------
      if ( $sEntryPointName eq 'online' && $sStandbyRole =~ /^master$/io ) {
        &logInfo( 20078,
        "<%s> STANDBY is Running on this node", $sMyName );
        $iReturnCode = SAPHotStandby::STANDBY;
      }
      elsif ( $sEntryPointName eq 'online' && $sStandbyRole =~ /^standby$/io ) {
        $iReturnCode = SAPHotStandby::ONLINE;
      }
      elsif ( $sStandbyRole =~ /^master$/io ) {
        $iReturnCode = VCSDef_MonitorExit::OFFLINE;
      }
      elsif ( $sStandbyRole =~ /^standby$/io ) {
        $iReturnCode = VCSDef_MonitorExit::ONLINE;
      }
    }
    elsif ( $lsCmdout [-1] eq 'ADMIN' ) {
      &logTrace ( 20079,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    elsif ( $lsCmdout [-1] eq 'STARTING' ) {
      &logTrace ( 20080,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    elsif ( $lsCmdout [-1] eq 'OFFLINE' ) {
      &logError ( 20081,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    elsif ( $lsCmdout [-1] eq 'SHUTDOWN' ) {
      &logError ( 20082,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    elsif ( $lsCmdout [-1] eq 'KILL' ) {
      &logError ( 20083,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    else {
      &logError ( 20084,
      "<%s> Database state returned undefined output [%s].",
      $sMyName, join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    }
  }
  elsif( ! defined $iExitCode ) {
    &logError ( 20085,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  else {
    &logError ( 20086,
    "<%s> Database state command returned invalid exit code [%s] and Cmd output [%s]",
    $sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String LCName,
                               IN String LCUser,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sLCName                   Name of the liveCache/MaxDB instance.
 $sLCUser                   Operating system user for liveCache/MaxDB server.
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sLCName,
       $sLCUser,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20087, "Subroutine <%s> called with:
  OnlinePids               [%s]
  LCName                   [%s]
  LCUser                   [%s]
  UseFFLM                  [%s]",

  $sMyName, $rliOnlinePids, $sLCName, $sLCUser, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  $sLCUser = "sdb"; # Change the user to liveCache process owner..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();
  my %hiLCPid = ();
  my $iPPid = ();

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == ( $iUid = getpwnam $sLCUser ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                  # And called to kill-by-UID..
    &logTrace ( 20088, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sLCUser, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( SAPHotStandby::GetSrvrDetails ( $sLCName, 
                                         \%hsSrvrDetails ) ) {
      &logTrace ( 20089,
      "<%s> Got details for monitoring SAPHotStandby instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20090,
      "<%s> Failed to get details for monitoring SAPHotStandby instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nSAPHotStandby process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = SAPHotStandby::FFLMStatus ( $sPidFile,
                                                $sProc,
                                                $iUid,
                                                \$iPid,
                                                \$iPPid );
          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
          $hiProcPids { $sKey } = $iPid;   # Store..

          $hiLCPid { $iPid } = $iPPid;    # by vchappav ...
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20091,
      "<%s> SAPHotStandby instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20092,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sLCUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20093,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20094,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20095,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      &logTrace ( 20096, "<%s> Searching PID [%s] UID [%s] CMD [%s]",
      $sMyName, $$rhlProc { PID }, $$rhlProc { UID },
      $$rhlProc { COMMAND } );

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {        # Now found..
                unshift ( @$rliOnlinePids, $iPid );             # Store..
                if ( VCSDef_MonitorExit::OFFLINE == $iFound &&  # Earlier marked offline..
                     ( ! defined $hiLCPid { $iPid } ) ) {       # Not already matched..    

                  $hiLCPid { $iPid } = $$rhlProc { PPID };         # Store PPID of the process..

                  &logTrace ( 20097,
                  "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                  $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                  $sProc );

                  $hiProcPids { $sKey } = $iPid;        # Store..
                  #unshift ( @$rliOnlinePids, $iPid );   # Store..
                  $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..

                  &logTrace ( 20098,
                  "<%s> Got [%s] processes into hiLCPid", $sMyName, scalar(keys(%hiLCPid)) );

                  #----------------
                  # Update status..
                  #----------------
                  push ( @lsUpdates, $sPidFile, $sProc, $iFound );
                  @{ $hsCopy { $sKey } } = @lsUpdates; # Update..

                  last;
                }
              }
            }
          }
        }
        else { # Request for kill..
          &logTrace ( 20099,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..

          $hiLCPid { $iPid } = $$rhlProc { PPID };         # Store PPID of the process..
        } 
      }
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20100,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( SAPHotStandby::CreatePidFile ( $sPidFile,
                                            $hiProcPids { $sKey } ) ) {
          &logTrace ( 20101,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20102,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound && ( 2 <= scalar (keys (%hiLCPid)) ) ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20103,
    "<%s> Found [%s] PIDs [%s] for this instance of SAPHotStandby.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20104,
    "<%s> SAPHotStandby process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  # Sort the PIDs of the HotStandby processes in the descending order of their parent PIDs.
  if( (scalar @$rliOnlinePids > 0) && (scalar @$rliOnlinePids == scalar (keys (%hiLCPid)) ) ) {
    &logTrace ( 20105,
    "<%s> Sorting the HotStandby process IDs in the descending order of their parent IDs.",
    $sMyName );

    @$rliOnlinePids = sort {$hiLCPid{$b} <=> $hiLCPid{$a} } keys %hiLCPid ;
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String EnvFile,
                          IN String LCUser,
                          IN String MonitorProgram );

 Description: Performs the in-depth monitoring

 Parameters :

 $sEnvFile                  Environments file for 'LCUser'. 
 $sLCUser                   Operating system user for liveCache/MaxDB server.
 $sMonitorProgram           Full path to external monitor program.
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sLCUser,
       $sMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20106, "Subroutine <%s> called with:
  EnvFile                  [%s]
  LCUser                   [%s]
  MonitorProgram           [%s]",

  $sMyName,  $sEnvFile, $sLCUser, $sMonitorProgram );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen = $^T + SAPHotStandby::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         SAPHotStandby::ExternalMonitor ( $sLCUser,
                                          $sEnvFile,
                                          $sExtCmd,
                                          $sExtCmdArgs,
                                          $iTillWhen ) ) ) {
      &logTrace ( 20107,
      "<%s> SAPHotStandby instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20108,
      "<%s> SAPHotStandby instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String DBMPassord,
                                        IN String DBMUserId,
                                        IN String EnvFile,
                                        IN String LCHome,
                                        IN String LCHost,
                                        IN String LCUser );

 Description : Just In Time Validation

 Parameters :

 $sDBMPassword              Password for Database Manager operator user.
 $sDBMUserId                Username for Database Manager operator.
 $sEnvFile                  Environments file for 'LCUser'. 
 $sLCHome                   Home directory for liveCache server.
 $sLCHost                   Virtual hostname for liveCache server.
 $sLCUser                   Operating system user for liveCache/MaxDB server.
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBMPassword,
       $sDBMUserId,
       $sEnvFile,
       $sLCHome,
       $sLCHost,
       $sLCUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20109, "Subroutine <%s> called with:
  DBMPassword              [%s]
  DBMUserId                [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCHost                   [%s]
  LCUser                   [%s]",

  $sMyName, $sDBMPassword, $sDBMUserId, $sEnvFile, 
  $sLCHome, $sLCHost, $sLCUser );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( ! -f  $sEnvFile ) {
    &logError ( 20110, "<%s> [%s] is not a plain file",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------------------
  # Validate parameter LCHome..
  #-----------------------------
  if ( ! -d  $sLCHome ) {
    &logWarn ( 20111, "<%s> [%s] directory does not exist",
    $sMyName, $sLCHome );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------------
  # Validate parameter LCHost..
  #--------------------------
  if ( ! inet_aton $sLCHost ) {
      &logError ( 20112, "<%s> LCHost [%s] is an invalid Hostname/IPv4 address",
      $sMyName, $sLCHost );
      $iReturnCode = 0; # Failure..
  }

  #-----------------------------
  # Validate parameter LCUser..
  #-----------------------------
  if ( ! defined getpwnam $sLCUser ) {
    &logError ( 20113, "<%s> [%s] is not a valid UNIX login",
    $sMyName, $sLCUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String LCUser,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sLCUser                   Operating system user for liveCache/MaxDB server.
 $sEnvFile                  Environments file for 'LCUser'. 
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sLCUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20114, "Subroutine <%s> called with:
  LCUser                   [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sLCUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdOut ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20115,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20116,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20117,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20118,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  SAPHotStandby::ExecLocalCmd ( $sLCUser, 
                                $sEnvFile, 
                                $sCmd, 
                                $sCmdArgs, 
                                \$iExitCode, 
                                $iTimeout, 
                                \@lsCmdOut );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    &logError ( 20119,
    "<%s> Command [%s] with args [%s] timed out",
    $sMyName, $sCmd, $sCmdArgs );
    $iExitCode = 'undef';
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20120,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }
  else {
    &logWarn ( 20121,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20122, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20123, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20124,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20125,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20126,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid,
       $riPPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20127, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20128, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20129, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20130, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20131,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20132,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20133,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( SAPHotStandby::GetPidDetails ( $iPid,
                                        \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20134,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20135,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid  = $iPid; # Store..
          $$riPPid = $$rhlProc { PPID }; # Store..  # by vchappav ...
        }
        else {
          &logTrace ( 20136,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20137,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20138, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool KillSAPHotStandbyProcesses ( IN String LCName,
                                              IN String LCUser );

 Description  : Kills processes for this instance of SAPHotStandby
 
 Parameters :

 $sLCName                 Name of the liveCache/MaxDB instance.
 $sLCUser                 Operating system user for liveCache/MaxDB server.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillSAPHotStandbyProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sLCName,
       $sLCUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20139, "Subroutine <%s> called with:
  LCName                   [%s]
  LCUser                   [%s]",

  $sMyName, $sLCName, $sLCUser );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'SAPHotStandby::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    SAPHotStandby::GetProcState ( \@liGoners,
                                  $sLCName,
                                  $sLCUser,
                                  0 ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  #@liGoners = sort { $a <=> $b } @liGoners;        # by vchappav...

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20140,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20141,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20142,
    "<%s> Could not locate any process this instance of SAPHotStandby",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference CmdOut ,
                                IN String HideString );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdOut        Stores o/p from command if successful
 $sHideString      String in CommandArgs to hide

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdOut ,
       $sHideString ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20143, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdOut   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdOut );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  my ( $iExitCode, $sDispStr, $sCmdDeCr, $sDeCrPasswd ) = ();

  #---------------------
  # Set up for display..
  #---------------------
  $sDispStr = $sCommandArgs;

  #-------------------------
  # Get decrypted password..
  #-------------------------
  if ( defined $sHideString ) {
    #-----------
    # Validate..
    #-----------
    if ( ! -x $sCmdVcsDecrypt ) {
      &logError ( 20144,
      "<%s> Command [%s] is not executable", $sMyName, $sCmdVcsDecrypt );
      return 0; # Failure..
    }

    $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
    $sDeCrPasswd = readpipe ( "$sCmdDeCr $sHideString" ); # Get decrypted..

    #--------------------
    # Collect exit code..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20145, "<%s> Failed to decrypt crypted string [%s]",
      $sMyName, $sHideString );
      return 0; # Failure..
    }

    $sCommandArgs =~ s/$sHideString/$sDeCrPasswd/;
    $sDeCrPasswd  =~ s/\$/./g;
  }

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdOut = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20146,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20147,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20148,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sDispStr );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20149,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sDispStr );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdOut,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           undef,
                                                           undef,
                                                           $sDeCrPasswd );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdOut;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20150,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20151,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
    }
  }
  else { # Execution error..
    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20152,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sDispStr, $iTimeout );
      $iReturnCode = 0; # Failure..
    }
    else { # Failure due to something else..
      &logError ( 20153,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
      $iReturnCode = 0; # Failure..
    }
    @$rlsCmdOut = (); # Reset..
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20154, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20155,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  if ( "solaris" eq $sSwitch ) {

      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

  }
  elsif ( "hpux" eq $sSwitch ) {
      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

  }
  elsif ( "aix" eq $sSwitch ) {
      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

  }
  elsif ( "linux" eq $sSwitch ) {
      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

  }

  &logTrace ( 20156, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20157,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20158,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String LCName,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sLCName                   Name of the liveCache/MaxDB instance.
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sLCName,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20159, "Subroutine <%s> called with:
  LCName                   [%s]
  SrvrDetails              [%s]",

  $sMyName, $sLCName, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat ) = ();

  #-------------------------------------
  # List of search patterns to monitor..
  #-------------------------------------
  my @lsSrchPatterns = ( qr|kernel $sLCName|,
                         qr|kernel $sLCName| );

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20160, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20161, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsSrchPatterns ) {
    $sKey = 'srchstr_' . ++ $iPat;
    $sPidFile = "$sPIDDir/pidfile.$sKey";
    @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20162, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  $sResourceType = "SAPHotStandby" if $sResourceType !~ /\w/; # Ensure..

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #-----------------------------------------
  # Form query string based on entry point..
  #-----------------------------------------
  $sCommand = $sCmdHatype; # Get global..
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
  $sCommandArgs = "-value $sResourceType $sAttribute";

  #----------------------------------
  # Fire command and collect output..
  #----------------------------------
  ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20163,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    $iEngineTimeout = $bUseMargin?$hiGraceMargin{$sAttribute}:0;
  }
  elsif ( ! $iEngineTimeout ) { # Empty value..
    &logTrace ( 20164, "<%s> Could not get [%s] from engine",
    $sMyName, $sAttribute );
    $iEngineTimeout = $bUseMargin?$hiGraceMargin{$sAttribute}:0;
  }
  else { # Success..
    &logTrace ( 20165, "<%s> Got [%s] as [%s] sec from engine",
    $sMyName, $sAttribute, $iEngineTimeout );
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                 # Margin requested..
       $iEngineTimeout >= $hiGraceMargin{$sAttribute} ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin{$sAttribute};       # Apply..

    &logTrace ( 20166,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin{$sAttribute}, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}


=pod

 -------------------------------------------------------------------------
 Function : int RemoveIPCResourcesForSAPHotStandby ( IN String User,
                                                     IN String LCHome, 
                                                     IN String LCName )

 Description : Removes IPC resources for SAP liveCache.
 
 Parameters  :

 $sUser        User name for which IPC resources needs to be removed.
 $sLCHome      The directory where dbmcli is located.
 $sLCName      Name of the liveCache instance for which IPCs needs to be removed.

 Returns:  
               boolean value in the set of {0,1} where
               0 Success
               1 failure
 -------------------------------------------------------------------------

=cut

sub RemoveIPCResourcesForSAPHotStandby {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sLCHome,
       $sLCName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20167, "Subroutine <%s> called with:
  User           [%s]
  LCHome         [%s]
  LCName         [%s]",
  $sMyName, $sUser, $sLCHome, $sLCName );

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPHotStandby::GetEngineTimeout ( 1 );

  #-------------------------------------------------------------
  # Try to get independant data path for SAP liveCache Server..
  #-------------------------------------------------------------
  my $sCommand = "$sLCHome/dbmcli";
  my $sCommandArgs = " dbm_getpath indepdatapath";

  my ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  my ( $sIndepDataPath, @liIPCshm, @liIPCsem, $sDbDir, $sUsDir, @lsDbFiles, @lsUsFiles ) = ();
  my $iReturnCode = 1;

  &logTrace ( 20168,
  "<%s> Getting independent data path for liveCache instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20169,
    "<%s> Insufficient time to get indepdatapath for liveCache instance. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    SAPHotStandby::ExecLocalCmd ( "root",
                                  "",
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20170,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logError ( 20171,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]. Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );
    $iReturnCode = 1;
  }
  if( scalar @lsCmdout ) {
    $sIndepDataPath = $lsCmdout[-1];
    chomp $sIndepDataPath;
  }
  else {
    $sIndepDataPath = "/sapdb/data" ;
    &logError(20172,
              "<%s> Failed to find the indepdatapath with the command [%s] with arguments [%s]",
              $sMyName, $sCommand, $sCommandArgs );
    &logInfo (20173,
              "<%s> Defaulting the indepdatapath for liveCache server [%s] to [%s]",
              $sMyName, $sLCName, $sIndepDataPath );
  }

  $sDbDir = $sIndepDataPath . "/ipc/" . "db\:" . $sLCName;
  $sUsDir = $sIndepDataPath . "/ipc/" . "us\:" . $sLCName;

  if( ReadDirectory( $sDbDir,
                     \@lsDbFiles ) ) {  # Read from db:<LCName> directory ...
    &logTrace(20174,
              "<%s> Read files [%s] from directory [%s]",
              $sMyName, join(' ', @lsDbFiles), $sDbDir );
    if( scalar @lsDbFiles && SAPHotStandby::GetIPCIDsFromFiles ( \@lsDbFiles,
                                                                 O_IPCSEM,
                                                                 \@liIPCsem ) ) {
      &logTrace(20175,
                "<%s> Read semaphore IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCsem), $sDbDir );
    }
    else {
      &logTrace(20176,
               "<%s> No semaphore IDs to read from the files in direcotory[%s]",
               $sMyName, $sDbDir );
    }

    if( scalar @lsDbFiles && SAPHotStandby::GetIPCIDsFromFiles ( \@lsDbFiles,
                                                                 O_IPCSHM,
                                                                 \@liIPCshm ) ) {
      &logTrace(20177,
                "<%s> Read Shared memory IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCshm), $sDbDir );
    }
    else {
      &logTrace(20178,
                "<%s> No Shared Memory IDs to read from the files in direcotory[%s]",
                $sMyName, $sDbDir );
    }
  }
  else {
    &logTrace( 20179,
               "<%s> Skiping directory [%s] from reading for IPC.",
               $sMyName, $sDbDir );
  }
  if( SAPHotStandby::ReadDirectory( $sUsDir,
                                    \@lsUsFiles ) ) {  # Read from us:<LCName> directory ...
    &logTrace(20180,
              "<%s> Read files [%s] from directory [%s]",
              $sMyName, join(' ', @lsUsFiles), $sUsDir );
    if( scalar @lsUsFiles && SAPHotStandby::GetIPCIDsFromFiles ( \@lsUsFiles,
                                                                 O_IPCSEM,
                                                                 \@liIPCsem ) ) {
      &logTrace(20181,
                "<%s> Read semaphore IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCsem), $sUsDir );
    }
    else {
      &logTrace(20182,
                "<%s> Failed to read to semaphore IDs from the files in direcotory[%s]",
                $sMyName, $sUsDir );
    }

    if( scalar @lsUsFiles && SAPHotStandby::GetIPCIDsFromFiles ( \@lsUsFiles,
                                                                 O_IPCSHM,
                                                                 \@liIPCshm ) ) {
      &logTrace(20183,
                "<%s> Read Shared memory IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCshm), $sUsDir );
    }
    else {
      &logTrace(20184,
               "<%s> Failed to read to Shared Memory IDs from the files in direcotory[%s]",
               $sMyName, $sUsDir );
    }
  }
  else {
    &logTrace( 20185,
              "<%s> Failed to read directory [%s].",
              $sMyName, $sUsDir );
  }
  if( scalar @liIPCsem ) {  # Check Semaphores for removing ...
    if( $iReturnCode = SAPHotStandby::RemoveIPCResourcesForUser ( $sUser,
                                                                  O_IPCSEM,
                                                                  \@liIPCsem ) ) {
      &logTrace (20186,
                 "<%s> Successfully removed Semaphores IDs [%s] for User [%s]",
                 $sMyName, join(' ', @liIPCsem), $sUser );
    }
    else {
      &logError (20187,
                 "<%s> Failed to remove Semaphores for User [%s]",
                 $sMyName, $sUser );
      $iReturnCode = 0; # Failure ...
    }
  }

  if( scalar @liIPCshm ) {  # Check Shared Memory for removing ...
    if( $iReturnCode = SAPHotStandby::RemoveIPCResourcesForUser ( $sUser,
                                                                  O_IPCSHM,
                                                                  \@liIPCshm ) ) {
      &logTrace (20188,
                 "<%s> Successfully removed Shared Memory IDs [%s] for User [%s]",
                 $sMyName, join(' ', @liIPCshm), $sUser );
    }
    else {
      &logError (20189,
                 "<%s> Failed to remove Shared Memory IDs for User [%s]",
                 $sMyName, $sUser );

      $iReturnCode = 0; # Failure ...
    }
  }
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ReadDirectory ( IN String Dir,
                                OUT ArrayOfStrings Files )

 Description : Reads the files in the directory given.
 
 Parameters  :

 $sDir         Directory from which the files needs to be read.
 $rlsFiles     Files in the directory passed.

 Returns:  
               boolean value in the set of {0,1} where
               0 Success
               1 failure
 -------------------------------------------------------------------------

=cut

sub ReadDirectory {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDir,
       $rlsFiles ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  &logTrace( 20190, "Subroutine <%s> called with:
  Dir            [%s]
  rlsFiles       [%s]",
  $sMyName, $sDir, $rlsFiles );

  if( ! -d $sDir ) {
    &logTrace ( 20191,
    "<%s> Directory [%s] is not valid. Skipping ..",
    $sMyName, $sDir );

    return 0;  # Failure ...
  }

  my ( $sFile, $sPwd, $sCurr ) = ();

  if ( opendir ( DIRENT, $sDir ) ) {
    &logTrace( 20192,
    "<%s> Directory [%s] opened for reading.",
    $sMyName, $sDir );

    @$rlsFiles = readdir ( DIRENT );

    closedir ( DIRENT ) ||
    &logWarn(20193,
    "<%s> Failed to close directory [%s] after reading",
    $sMyName, $sDir );
  }
  else {
    &logWarn(20194,
    "<%s> Failed to open Directory [%s] for reading.",
    $sMyName, $sDir );

    return 0;  # Failure ...
  }

  return 1;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetIPCIDsFromFiles ( IN ArrayOfString Files,
                                     IN Integer IPCFlags,
                                     OUT ArrayOfIntegers IPCs )

 Description : Extracts IPC IDs from the IPC files passed in.
 
 Parameters  :

 $rlsFiles     List of IPC file names from which IDs are extracted.
 $riIPCFlags   IPC flags indicating the type of the IPC IDs need to be extracted.
 $rliIPCs      List of IPC IDs which are extracted from the IPC files passed in.

 Returns:  
               boolean value in the set of {0,1} where
               0 Success
               1 failure
 -------------------------------------------------------------------------

=cut
sub GetIPCIDsFromFiles {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rlsFiles,
       $riIPCFlags,
       $rliIPCs ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20195, "Subroutine <%s> called with:
  rlsFiles       [%s]
  IPC Flags      [%s]
  rliIPCs        [%s]",
  $sMyName, $rlsFiles, $riIPCFlags, $rliIPCs );

  foreach my $sFile (@$rlsFiles) {
    chomp $sFile;
    my $iIPC = $sFile;

    if( O_IPCSHM == $riIPCFlags && $iIPC =~ s/^-m//o ) {
      &logTrace(20196,
      "<%s> Got Shared Memory ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
    elsif( O_IPCSEM == $riIPCFlags && $iIPC =~ s/^-s//o ) {
      &logTrace(20197,
      "<%s> Got Semaphore ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
    elsif( O_IPCMSQ == $riIPCFlags && $iIPC =~ s/^-q//o ) {
      &logTrace(20198,
      "<%s> Got Message Queue ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
    elsif( O_IPCALL == $riIPCFlags && $iIPC =~ s/^-[sqm]+//o ) {
      &logTrace(20199,
      "<%s> Got IPC ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
  }

  return 1;
}

=pod
 -------------------------------------------------------------------------
 Function : bool RemoveIPCResourcesForUser ( IN String User,
                                             IN Integer IpcFlags,
                                             IN RefListInteger Ids );

 Description : Removes IPC resources for specified UNIX user

 Parameters :

 $sUser      UNIX user whose IPC resources are to be removed
 $iIpcFlags  Behaviour flag
 $rliIds     List of IPC IDs to remove for that user

 Values for Behaviour flag:
 O_IPCSHM : Shared Memory           O_IPCSEM : Semaphore
 O_IPCMSQ : Message Queue           O_IPCALL : All of the above

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub RemoveIPCResourcesForUser {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $iIpcFlags,
       $rliIds ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  &logTrace( 20200, "Subroutine <%s> called with:
  User           [%s]
  IPC Flags      [%s]
  rliIds         [%s]",
  $sMyName, $sUser, $iIpcFlags, $rliIds );

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sFormat = 'l l l l ul ul i'; # Needed for unpack. Check intro(2)..
  my $iLocn = $^O eq 'linux'?1:0;  # Index locn. of ipc_perm.uid..
  my $lsVar = 0;                   # Stores the binary structure semid_ds..
  my $sCommand = $sCmdIpcs;        # Get global..

  my ( $iId, $iUid, $iIpcUid, $iRemoved,
       @liShmIds, @liMqIds, @liSemIds, @liIdsRemoved, %hiIds ) = ();

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;


  #--------------------
  # Validate the user..
  #--------------------
  unless ( defined ( $iUid = getpwnam $sUser ) ) {
    &logError ( 20201, "<%s> Invalid User [%s] specfied. Cannot proceed.",
    $sMyName, $sUser );
    return 0; # Error..
  }

  #----------------
  # Check for IDs..
  #----------------
  if ( scalar @$rliIds ) {
    &logTrace ( 20202,
    "<%s> Request to remove only [%s] IDs for User [%s]",
    $sMyName, join ( ' ', @$rliIds ), $sUser );

    %hiIds = map { $_ => 1 } @$rliIds; # Remove duplicates..
  }

  #-------------------------------
  # Read in the contents of IPCs..
  #-------------------------------
  foreach ( readpipe ( $sCommand ) ) {
    s/\n//; # Chomp..

    #---------------
    # Handle Linux..
    #---------------
    if ( $^O eq 'linux' ) {
      if ( $iIpcFlags & O_IPCSHM && # Shared Memory..
           /shmid/ ... /^$/ ) {
        /(\w+)\s+(\d+)/;
        push ( @liShmIds, $2 ) if defined $2;
      }

      if ( $iIpcFlags & O_IPCSEM && # Semaphore..
           /semid/ ... /^$/ ) {
        /(\w+)\s+(\d+)/;
        push ( @liSemIds, $2 ) if defined $2;
      }

      if ( $iIpcFlags & O_IPCMSQ && # Message Queue..
           /msqid/ ... /^$/ ) {
        /(\w+)\s+(\d+)/;
        push ( @liMqIds, $2 ) if defined $2;
      }
    }
    else { # the rest..
      #--------------------
      # Skip removed IPCs..
      #--------------------
      if ( /^[m|s|q]\s*(\d+)\s+\w+\s+(\S)/ && # Of all three..
           defined $2                      && # Mode defined..
           $2 eq 'D') {                       # Says, it's removed..
        next;                                 # Skip..
      }

      $iIpcFlags & O_IPCSHM && # Shared Memory..
      push ( @liShmIds, $_ =~ /^m\s*(\d+).*/ );

      $iIpcFlags & O_IPCMSQ && # Message Queue..
      push ( @liMqIds,  $_ =~ /^q\s*(\d+).*/ );

      $iIpcFlags & O_IPCSEM && # Semaphore..
      push ( @liSemIds, $_ =~ /^s\s*(\d+).*/ );
    }
  }
  #---------------------------------------------
  # Process each one separately as per request..
  #---------------------------------------------

  #----------------
  # Shared Memory..
  #----------------
  if ( $iIpcFlags & O_IPCSHM ) {
    foreach $iId ( @liShmIds ) {

      next if ( scalar @$rliIds && ! $hiIds { $iId } );

      shmctl( $iId, IPC_STAT, $lsVar ) ||
      &logError ( 20203,
      "<%s> Failed to use shmctl(2) on SHM Id [%s]", $sMyName, $iId );

      #-----------------------------------------------------
      # Unpack uid(2) out of the packed semid_ds structure..
      #-----------------------------------------------------
      $iIpcUid = ( unpack ( $sFormat, $lsVar ) ) [$iLocn];

      #-----------------------------
      # Remove resource if a match..
      #-----------------------------
      if ( $iIpcUid == $iUid ) {
        if ( shmctl ( $iId, IPC_RMID, 0 ) ) {
          push ( @liIdsRemoved, $iId );
          ++$iRemoved; # Increment..
        }
        else {
          &logError ( 20204,
          "<%s> Failed to remove SHM Id [%s]. Error [%s]", $sMyName, $iId, $! );
          $iReturnCode = 0; # Failed..
        }
      }
    }
  }
  if ( scalar @liIdsRemoved ) {
    &logTrace ( 20205, "<%s>
    +--- Removed SHM Ids: [%s] for User [%s] Uid [%s]. Total [%s] ---+",
    $sMyName, join ( ' ', @liIdsRemoved ), $sUser, $iUid, scalar @liIdsRemoved );
    @liIdsRemoved = (); # Clear..
  }

  #----------------
  # Message Queue..
  #----------------
  $lsVar = 0; # Reset..
  if ( $iIpcFlags & O_IPCMSQ ) {
    foreach $iId ( @liMqIds ) {

      next if ( scalar @$rliIds && ! $hiIds { $iId } );

      msgctl( $iId, IPC_STAT, $lsVar ) ||
      &logError ( 20206,
      "<%s> Failed to use msgctl(2) on SHM Id [%s]", $sMyName, $iId );

      #-----------------------------------------------------
      # Unpack uid(2) out of the packed semid_ds structure..
      #-----------------------------------------------------
      $iIpcUid = ( unpack ( $sFormat, $lsVar ) ) [$iLocn];

      #-----------------------------
      # Remove resource if a match..
      #-----------------------------
      if ( $iIpcUid == $iUid ) {
        if ( msgctl ( $iId, IPC_RMID, 0 ) ) {
          push ( @liIdsRemoved, $iId );
          ++$iRemoved; # Increment..
        }
        else {
          &logError ( 20207,
          "<%s> Failed to remove Msg QId [%s]. Error [%s]", $sMyName, $iId, $! );
          $iReturnCode = 0; # Failed..
        }
      }
    }
  }

  if ( scalar @liIdsRemoved ) {
    &logTrace ( 20208, "<%s>
    +--- Removed Msg QIds: [%s] for User [%s] Uid [%s]. Total [%s] ---+",
    $sMyName, join ( ' ', @liIdsRemoved ), $sUser, $iUid, scalar @liIdsRemoved );
    @liIdsRemoved = (); # Clear..
  }


  #------------
  # Semaphore..
  #------------
  $lsVar = 0; # Reset..
  if ( $iIpcFlags & O_IPCSEM ) {
    foreach $iId ( @liSemIds ) {

      next if ( scalar @$rliIds && ! $hiIds { $iId } );

      semctl( $iId, 0, IPC_STAT, $lsVar ) ||
      &logError ( 20209,
      "<%s> Failed to use semctl(2) on Sem Id [%s]", $sMyName, $iId );

      #-----------------------------------------------------
      # Unpack uid(2) out of the packed semid_ds structure..
      #-----------------------------------------------------
      $iIpcUid = ( unpack ( $sFormat, $lsVar ) ) [$iLocn];

      #-----------------------------
      # Remove resource if a match..
      #-----------------------------
      if ( $iIpcUid == $iUid ) {
        if ( semctl ( $iId, 0, IPC_RMID, 0 ) ) {
          push ( @liIdsRemoved, $iId );
          ++$iRemoved; # Increment..
        }
        else {
          &logError (20210,
          "<%s> Failed to remove Sem Id [%s]. Error [%s]", $sMyName, $iId, $! );
          $iReturnCode = 0; # Failed..
        }
      }
    }
  }

  if ( scalar @liIdsRemoved ) {
    &logTrace ( 20211,
    "<%s>
    +--- Removed Sem Ids: [%s] for User [%s] Uid [%s]. Total [%s] ---+",
    $sMyName, join ( ' ', @liIdsRemoved ), $sUser, $iUid, scalar @liIdsRemoved );
    @liIdsRemoved = (); # Clear..
  }

  #---------------------------------
  # Log, if we did anything useful..
  #---------------------------------
  if ( $iRemoved ) {
    &logTrace( 20212,
    "<%s> Successfully removed [%s] IPC resources for User [%s] in all.",
    $sMyName, $iRemoved, $sUser );
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool OnlineStandby ( IN String Sys,
                                 IN String Grp,
                                 IN String Reason,
                                 IN String FaultedNode );

 Description : Pre-online trigger for SAPHotStandby

 Parameters :

 $sSys                System where the group is asked to be made online
 $sGrp                Group for which preonline has been invoked
 $sReason             Contains the string MANUAL or FAULT
 $sFaultedNode        Cluster node where the fault occurred due to FAULT

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub OnlineStandby {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSys,
       $sGrp,
       $sReason,
       $sFaultedNode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20213, "Subroutine <%s> called with:
  Sys             [%s]
  Grp             [%s]
  Reason          [%s]
  FaultedNode     [%s]",

  $sMyName,
  $sSys, $sGrp, $sReason, $sFaultedNode );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHares; # Get global..

  my %hsInstMatch = ( 'STANDBY'  => 'MASTER' );

  my ( $i,  $iExitCode,  $sRes,  $sGroup,  $sNodeName,   $sMasterRes,
       $sCmdArgs,  $sTempNode,   $sStandbyRes,  $sStandbyRole,   $sFailHost,
       $sKey,  $sLCName,  @lsHaresOut,    @lsCpy,      @lsValidNodes,
       @lsStandbys,      @lsMasters,    %hsMasters,        %hsStandbys,
       %hsStandbyGrps, %hsStandbyHosts, %hsSwStandbys, %hsOnlineMasters ) = ();

  #---------------------
  # Validate arguments..
  #---------------------
  if ( ! inet_aton ( $sSys ) ) {
    &logError ( 20214,
    "<%s> Destintation cluster node [%s] cannot be resolved",
    $sMyName, $sSys );
    $iReturnCode = 0; # Error..
  }
  elsif ( $sReason eq 'MANUAL' && ! inet_aton ( $sFaultedNode ) ) { # Manual online
    &logInfo ( 20215,
    "<%s> Faulted cluster node [%s] cannot be resolved. Invoked during manual online",
    $sMyName, $sFaultedNode );
    #return 0; # Do nothing ..
  }
  elsif ( $sReason eq 'MANUAL' ) {  # Manual switchover 
    &logInfo ( 20216,
    "<%s> Invoked during a normal switch-over.", $sMyName );
    #$iReturnCode = 0; # Do nothing..
  }
  elsif ( $sReason eq 'FAULT' && ! inet_aton ( $sFaultedNode ) ) { # System reboot/shutdown 
    &logInfo ( 20217,
    "<%s> Faulted cluster node [%s] cannot be resolved. Continueing ...",
    $sMyName, $sFaultedNode );
    #$iReturnCode = 0; # Error..
  }
  elsif ( ! inet_aton ( $sFaultedNode ) ) { # Normal Failover ...
    &logError ( 20218,
    "<%s> Faulted cluster node [%s] cannot be resolved.",
    $sMyName, $sFaultedNode );
    $iReturnCode = 0; # Error..
  }
  elsif ( ! -e $sCmd ) {
    &logError ( 20219,
    "<%s> Command [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Error..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20220,
    "<%s> Command [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Error..
  }

  #$iReturnCode || return $iReturnCode; # Return, if failed..
  $iReturnCode || return 1; # Return, if failed..

  #----------------
  # Re-initialise..
  #----------------
  $sCmdArgs = "-display -type SAPHotStandby -attribute " .
              "State Group StandbyRole MasterResName LCName";

  #---------------------------------------------
  # Assume the group name is valid and proceed..
  #---------------------------------------------
  @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sCmd $sCmdArgs" );

  #-----------------------------------------
  # Check if command executed successfully..
  #-----------------------------------------
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20221,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    return 0; # Error..
  }
  elsif ( $lsHaresOut[0] =~ /10554.*No resource exists with/o ) {
    &logError ( 20222,
    "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
    $sMyName, $sCmd, $sCmdArgs,  $lsHaresOut[0] );
    return 0; # Error..
  }

  $sFailHost = $sSys;    # Set to default..
  push @lsHaresOut, "#"; # Insert trailing de-marker..
  @lsCpy = @lsHaresOut;  # Backup..

  #---------------------------------------------------
  # Mark all MASTER/STANDBY resources in given group..
  #---------------------------------------------------
  @lsMasters =
  grep {
     ( s/\n// || 1 )                   && # Chomp, and always make true..
     ( /Group.*\s+(\S+)/               && # Mark the group name..
     ( $sGroup = $1 ) && 0 )           || # Store, and always make false..
     ( s/.*LCName.*\s+// )             || # Get the LCName..
     ( s/^(\S+).*StandbyRole.*\s          # Mark resource..
     (MASTER|STANDBY)/$1:$2/ix         && # If MASTER or STANDBY variant..
     ++ $i )                           || # Set flag..
     ( $i                              && # If flag..
     s/#/sprintf ( $sGroup, -- $i )/e )   # Print group.. reset flag..
  } @lsCpy;

  &logTrace ( 20223,
  "<%s> Searching MASTER resource in group [%s]", $sMyName, $sGrp );

  &logTrace ( 20224,
  "<%s> All MASTER resources [%s]", $sMyName, join ( ' ', @lsMasters ) );

  #-------------------
  # Start processing..
  #-------------------
  while ( 3 == ( ( $sLCName, $sRes, $sGroup ) = 
    splice ( @lsMasters, 0, 3 ) ) ) {
    ( $sRes, $sStandbyRole ) = $sRes =~ /([^:]+)/g;

    #-------------------------------
    # Passed group has MASTER|STANDBY ..
    #-------------------------------
    if ( $sGrp eq $sGroup ) {
      if ( $sStandbyRole =~ /MASTER/io ) {
        &logInfo ( 20225,
        "<%s> +------- Group [%s] has MASTER variant -------+",
        $sMyName, $sGrp );

        #--------
        # Store..
        #--------
        push @{ $hsMasters { $sRes } },
        $sLCName, $sStandbyRole, $sGroup; # Store..
      }
      elsif ( $sStandbyRole =~ /STANDBY/io ) {
        &logInfo ( 20226,
        "<%s> +------- Group [%s] has STANDBY variant -------+",
        $sMyName, $sGrp );

        #--------
        # Store..
        #--------
        push @{ $hsSwStandbys { $sRes } },
        $sLCName, $sStandbyRole, $sGroup; # Store..
        $sStandbyRes = $sRes;
      }
    }

    &logTrace ( 20227,
    "<%s> Resource [%s] LCName [%s] StandbyRole [%s] found in Group [%s]",
    $sMyName, $sRes, $sLCName, $sStandbyRole, $sGroup );
  }

  #-----------------------------------------------
  # Find online resources and list their details..
  #-----------------------------------------------
  @lsCpy = @lsHaresOut; # Re-initialise..
  $i = 0;

  &logTrace ( 20228,
  "<%s> Looking for all ONLINE SAPHotStandby resources", $sMyName );

  @lsStandbys =
    grep {
      ( s/\n// || 1 )                        && # Chomp, and always make true..
      ( /^(\S+).*Group.*\s+(\S+)/            && # Mark resource, and group name..
      ( $sGroup = "$1:$2" ) && 0 )           || # Store, and always make false..
      ( s/.*\s(\S+)\s+ONLINE$/$1/ && ++ $i ) || # If online, get node name, set flag..
      ( $i                                   && # If flag..
      /MasterResName|StandbyRole|LCName/     && # and contains one of these..
      s/.*\s+// )                            || # get the last field..
      ( $i                                   && # If flag, use demarker..
      s/#/sprintf ( $sGroup, -- $i )/e )        # to print the group name.. reset flag..
    } @lsCpy;                                   # From the copy..

  &logTrace ( 20229,
  "<%s> All ONLINE SAPHotStandby resources [%s]", $sMyName, join( ' ', @lsStandbys) );

  while ( 5 == ( ( $sNodeName, $sLCName, $sMasterRes, $sStandbyRole, $sGroup ) =
    splice ( @lsStandbys, 0, 5 ) ) ) {
    ( $sRes, $sGroup ) = $sGroup =~ /([^:]+)/g;

     #--------------------------------------
     # Store only MASTER|STANDBY variants..
     #--------------------------------------
     if ( $sStandbyRole =~ /STANDBY/io && 
          scalar keys %hsMasters  && 
          ( $sLCName eq ${ $hsMasters { $sMasterRes } }[0] ) ) {
       &logTrace ( 20230, "<%s> Located STANDBY variant. Resource [%s] details:
       Group [%s] Node [%s] MasterResName [%s] StandbyRole [%s] LCName [%s]",
       $sMyName, $sRes, $sGroup, $sNodeName, $sMasterRes, $sStandbyRole, $sLCName );

       push @{ $hsStandbys { $sRes } },
       $sGroup, $sMasterRes, $sStandbyRole, $sLCName, $sNodeName;
     }
     elsif ( $sStandbyRole =~ /MASTER/io && 
             scalar keys %hsSwStandbys &&
             ( $sLCName eq ${ $hsSwStandbys { $sStandbyRes } }[0] ) ) {
       &logTrace ( 20231, "<%s> Located MASTER variant. Resource [%s] details:
       Group [%s] Node [%s] MasterResName [%s] StandbyRole [%s] LCName [%s]",
       $sMyName, $sRes, $sGroup, $sNodeName, $sMasterRes, $sStandbyRole, $sLCName );

       push @{ $hsOnlineMasters { $sRes } },
       $sGroup, $sMasterRes, $sStandbyRole, $sLCName, $sNodeName;
     }
  }

  #--------------------------------------------
  # Validate if group has MASTER|STANDBY variant..
  #--------------------------------------------
  @lsCpy = @lsHaresOut; # Re-initialise.

  if ( keys %hsMasters ) {
    &logInfo ( 20232,
    "<%s> Failover Group [%s] has MASTER variant resources [%s].",
    $sMyName, $sGrp, join ( ' ' , keys %hsMasters ) );
  }
  elsif ( keys %hsSwStandbys ) {
    #------------------------------------------
    # Get all valid nodes for the faulted STANDBY ..
    #------------------------------------------
    @lsCpy = @lsHaresOut;          # Re-initialise.
    ( $sRes ) = keys %hsSwStandbys; # Take only one..

    &logInfo ( 20233,
    "<%s> Failover Group [%s] has STANDBY variant resource [%s]",
    $sMyName, $sGrp, $sRes );

    @lsValidNodes =
    grep {
      ( s/\n// || 1 )                     &&  # Chomp, and always make true..
      s/^$sRes.*State\s+(\S+).*LINE/$1/ ||    # Pick the node not FAULTed..
      s/^$sRes.*MasterResName.*\s+(\S+)/::$1/ # And MasterResName..
    } @lsCpy;                                 # From cmd output..

    &logTrace ( 20234,
    "<%s> Valid STANDBY resource nodes and MasterResName for failover [%s].", 
    $sMyName, join( ' ', @lsValidNodes ) );

    #-------------------------------------------------
    # Now extract MasterResName and configured nodes..
    #-------------------------------------------------
    @lsCpy = @lsValidNodes; # Re-use..
    @lsValidNodes = ();     # Truncate..

    foreach $sKey ( @lsCpy ) {
      if ( $sKey =~ /^::/ ) { # MasterResName found..
        ( $sMasterRes = $sKey ) =~ s/:://;
      }
      else {
        push ( @lsValidNodes, $sKey );
      }
    }

    &logTrace ( 20235,
    "<%s> Valid STANDBY resource nodes for failover [%s].", 
    $sMyName, join( ' ', @lsValidNodes ) );

    #-----------------------------
    # Now validate MasterResName..
    #-----------------------------
    if ( scalar keys %hsOnlineMasters && defined $hsOnlineMasters { $sMasterRes } &&  # Defined..
         ( $hsInstMatch { ${ $hsSwStandbys { $sRes } } [1] } eq  # StandbyRole pair..
         ${ $hsOnlineMasters { $sMasterRes } } [2] )         &&  # is a match..
         ( ${ $hsSwStandbys { $sRes } } [0]                  eq  # LCName..
           ${ $hsOnlineMasters { $sMasterRes } } [-2] ) ) {      # matches..
      $sTempNode = ${ $hsOnlineMasters { $sMasterRes } } [-1];   # Get the online node..
      ( $sFailHost ) = grep {                                    # Find the node other than..
                         ! /^($sTempNode|$sFaultedNode)$/o       # Master online or Standby faulted..
                       } @lsValidNodes;                          # From all posibilities..

      #---------------------------------------------------
      # Check if we are in a position to failover safely..
      #---------------------------------------------------
      if ( defined $sFailHost ) {
        &logInfo ( 20236,
        "<%s> Failing over group [%s] to node [%s]", $sMyName, $sGrp, $sFailHost );
      }
      else { # Prevent onlining..
        &logError ( 20237,
        "<%s> Could not locate any node to online Group [%s]",
        $sMyName, $sGrp );
        return $iReturnCode; # Error.. Prevent online of STANDBY..
      }

      #------------------------------------------------------------------------------
      # Online group to node where Master is NOT online or Standby is not faulted..
      #------------------------------------------------------------------------------
      if ( SAPHotStandby::OnlineGrp ( $sGrp,
                                      $sFailHost ) ) {
        &logInfo ( 20238,
        "<%s> Issued online of Group [%s] on node [%s]",
        $sMyName, $sGrp, $sFailHost );
        return $iReturnCode; # Success..
      }
      else {
        &logError ( 20239,
        "<%s> Could not issue online of Group [%s] on node [%s]",
        $sMyName, $sGrp, $sFailHost );
        return 0; # Error..
      }
    }
    else {
      &logError ( 20240,
      "<%s> Failover group [%s] does not have any valid MasterResName Online. Prevent STANDBY online before MASTER.",
      $sMyName, $sGrp );
      return 1; # Do nothing..
    }
  }
  else {
    &logInfo ( 20241,
    "<%s> Failover group [%s] does not have any MASTER|STANDBY variant",
    $sMyName, join ( ' ', keys %hsMasters ), $sGrp );
    return 0; # Do nothing..
  }

  #------------------------------------------------
  # Validate if we have any online STANDBY variant..
  #------------------------------------------------
  if ( ! keys %hsStandbys && $sReason eq 'MANUAL' && ! inet_aton ( $sFaultedNode ) ) { # Manual Online
    $sFailHost = $sSys; # Fail-over to the default node..
    &logInfo ( 20242,
    "<%s> Onlining group [%s] on node [%s]", $sMyName, $sGrp, $sFailHost );

    #-------------------------------
    # Online group to default node..
    #-------------------------------
    if ( SAPHotStandby::OnlineGrp ( $sGrp,
                                    $sFailHost ) ) {
      &logInfo ( 20243,
      "<%s> Issued online of Group [%s] on node [%s]",
      $sMyName, $sGrp, $sFailHost );
      return $iReturnCode; # Success..
    }
    else {
      &logError ( 20244,
      "<%s> Could not issue online of Group [%s] on node [%s]",
      $sMyName, $sGrp, $sFailHost );
      return 0; # Error..
    }
  }
  elsif( ! keys %hsStandbys && $sReason eq 'FAULT' ) { 
    # Standby needs to be online for switch-over or system shutdown or normal  group fault.
    $sFailHost = $sSys; # Fail-over to the default node..
    &logCritical ( 20245,
    "<%s> No STANDBY servers found ONLINE in cluster.
    Cannot failover service group [%s] for Master on to node [%s]", $sMyName, $sGrp, $sFailHost );

    return 1;
  }

  &logTrace ( 20246,
  "<%s> +----- STANDBY online node and group details -----+", $sMyName );

  #----------------------------------------------
  # Remove STANDBYs whose StandbySrv aren't found..
  #----------------------------------------------
  foreach $sKey ( keys %hsStandbys ) {
    if ( ! $hsMasters { ${$hsStandbys { $sKey } }[1] } ) { # Not in passed group..
      &logTrace ( 20247,
      "<%s> MasterResName [%s] resource not found in Grp [%s].
      Marked resource [%s] invalid",
      $sMyName, ${$hsStandbys { $sKey } }[1], $sGrp, $sKey );

      delete $hsStandbys { $sKey };                        # Then delete..
    }
    else {
      ++ $hsStandbyHosts { ${$hsStandbys { $sKey } }[-1] }; # Else store online node..
      ++ $hsStandbyGrps  { ${$hsStandbys { $sKey } }[0] };  # Else store group..

      &logTrace ( 20248,
      "<%s> Online node [%s] Group [%s]",
      $sMyName, ${$hsStandbys{$sKey}}[-1], ${$hsStandbys{$sKey}}[0] );
    }
  }

  #-----------------------------------------------
  # Check if we have STANDBYs in different groups..
  #-----------------------------------------------

  # Need to handle for multiple Standbys in future based on delay time. --- by vchappav
  if ( keys %hsStandbyGrps > 1 ) { 
    &logCritical ( 20249,
    "<%s> STANDBYs configured in different service groups [%s]",
    $sMyName, join ( ' ', keys %hsStandbyGrps ) );
  }

  #----------------------------------------------
  # Now check if we have a unique failover node..
  #----------------------------------------------
  if ( keys %hsStandbyHosts > 1 ) {
    &logCritical ( 20250,
    "<%s> STANDBYs found online on different nodes [%s]",
    $sMyName, join ( ' ', keys %hsStandbyHosts ) );

    #------------------------------------------
    # Delete entries for all other than STANDBY..
    #------------------------------------------
    foreach $sKey ( keys %hsStandbys ) {
      if ( ${$hsStandbys { $sKey } }[2] ne 'STANDBY' ) {
        &logError ( 20251,
        "<%s> Marking resource [%s] with StandbyRole [%s] as invalid",
        $sMyName, $sKey, ${$hsStandbys { $sKey } }[2] );

        #--------------------------
        # Delete in correct order..
        #--------------------------
        delete $hsStandbyHosts { ${$hsStandbys { $sKey } }[-1] }; # Delete node first..
        delete $hsStandbyGrps  { ${$hsStandbys { $sKey } }[0] };  # Delete group next..
        delete $hsStandbys { $sKey };                             # Now the resource..
      }
    }
  }

  #-------------------------
  # Find out failover node..
  #-------------------------
  foreach $sKey ( keys %hsStandbys ) {
    &logTrace ( 20252,
    "<%s> Standby server [%s] details:
    Group [%s] MasterResName [%s] StandbyRole [%s] LCName [%s] online on [%s]",
    $sMyName, $sKey,
    ${ $hsStandbys { $sKey } }[0], ${ $hsStandbys { $sKey } }[1],
    ${ $hsStandbys { $sKey } }[2], ${ $hsStandbys { $sKey } }[3],
    ${ $hsStandbys { $sKey } }[4] );

    #-----------------------------------------------------------
    # Master has been defined in MasterResName of STANDBY variant..
    #-----------------------------------------------------------
    if ( scalar @{ $hsMasters { ${$hsStandbys { $sKey }}[1] } } ) { # MasterResName is correct..
      if ( ${ $hsMasters { ${$hsStandbys { $sKey }}[1] } }[0] ne    # LCName of MASTER variant..
                            ${$hsStandbys { $sKey }}[3] ) {         # ne LCName of STANDBY ...
        &logError ( 20253,
        "<%s> LCName [%s] of STANDBY resource [%s] differs from
        LCName [%s] of MASTER resource [%s]",
        $sMyName, ${$hsStandbys { $sKey }}[3], $sKey,
        ${ $hsMasters { ${$hsStandbys { $sKey }}[1] } }[0],
        ${ $hsStandbys { $sKey } }[1] );
      }
      elsif ( uc ${ $hsMasters { ${$hsStandbys { $sKey }}[1] } }[1] ne # MASTER variant..
              $hsInstMatch { uc ${$hsStandbys {$sKey }}[2] } ) {        # not correct match..
        &logError ( 20254,
        "<%s> StandbyRole [%s] of STANDBY resource [%s] not a match for
        StandbyRole [%s] of MASTER resource [%s]", $sMyName,
        ${$hsStandbys { $sKey }}[2], $sKey,
        ${ $hsMasters { ${$hsStandbys { $sKey }}[1] } }[1],
        ${ $hsStandbys { $sKey } }[1] );
      }
      else { # All okay..
        &logTrace ( 20255,
        "<%s> MasterResName [%s] matches with:
        MASTER LCName [%s] StandbyRole [%s] Group [%s]", $sMyName,
        ${$hsStandbys { $sKey }}[1],
        @{ $hsMasters { ${$hsStandbys { $sKey }}[1] } } );

        $sFailHost = ${ $hsStandbys { $sKey } }[-1]; # Fail-over to STANDBY's node..
      }
    }
    else { # MasterResName is wrong..
      &logError ( 20256,
      "<%s> MasterResName [%s] value is in-correct for STANDBY resource [%s]",
      $sMyName,  ${$hsStandbys { $sKey }}[1], $sKey );
    }
  }

  &logInfo ( 20257, "<%s> Failing over Group [%s] to node [%s]",
  $sMyName, $sGrp, $sFailHost );

  #-------------------------------
  # Online group to Standby node..
  #-------------------------------
  if ( SAPHotStandby::OnlineGrp ( $sGrp,
                                  $sFailHost ) ) {
    &logInfo ( 20258, "<%s> Issued online of Group [%s] on node [%s]",
    $sMyName, $sGrp, $sFailHost );
  }
  else {
    &logError ( 20259,
    "<%s> Could not issue online of Group [%s] on node [%s]",
    $sMyName, $sGrp, $sFailHost );
    $iReturnCode = 0; # Do nothing..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool OnlineGrp ( IN String Grp,
                             IN String Sys );

 Description : Pre-online trigger for SAPHotStandby

 Parameters :

 $Grp                 Cluster group to online
 $sSys                System where the group is asked to be made online

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub OnlineGrp {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sGrp,
       $sSys ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20260, "Subroutine <%s> called with:
  Grp             [%s]
  Sys             [%s]",

  $sMyName, $sGrp, $sSys );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHagrp; # Get global..
  my @lsCmdArgs = ( "-online", "-nopre", "$sGrp", "-sys", "$sSys" );
  my $iExitCode = undef;

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ! -e $sCmd ) {
    &logError ( 20261,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20262,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  else { # Execute..
    &logTrace ( 20263,
    "<%s> Executing Cmd [%s] with args [%s]",
    $sMyName, $sCmd, join ( ' ', @lsCmdArgs ) );

    system ( $sCmd, @lsCmdArgs ); # Execute..

    #-------------------------
    # Check execution status..
    #-------------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20264,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
      $sMyName, $sCmd, join ( ' ', @lsCmdArgs ), $iExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20265,
      "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCmd, join ( ' ', @lsCmdArgs ), $iExitCode );
    }
  }

  return $iReturnCode;
}


1; # So PERL knows its included..
